﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Messaging;
using System.Text;
using System.Threading.Tasks;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;
using DocumentFormat.OpenXml.Packaging;

namespace OpenXmlExample
{
    class DocxGenerationDaemon
    {
        static string GenerateDocumentsQueueLocalName = @".\PRIVATE$\GenerateDocuments";
        static MessageQueue GenerateDocumentsQueue;

        static void Main(string[] args)
        {
            Initialize();
            MessageLoop();
        }

        /*
         * Incoming message:
         * 
         * Message label: GenerateDocuments
         * 
         * <GenerateDocuments>
         *   <ReturnQueue Name=".\PRIVATE$\GenerateDocumentClient" />
         *   <TemplateDocument Name="C:\Users\Public\Documents\TemplateDocuments\TemplateDocument.docx" />
         *   <GeneratedDocumentsDirectory Name="C:\Users\Public\Documents\GeneratedDocuments" />
         *   <DocumentGenerationData>
         *     <!-- the XML here is specific to the template document.  If the template document changes, the schema of the following will change to match it. -->
         *     <Customers>
         *       <Customer>
         *         <CustomerID>1</CustomerID>
         *         <Name>Cheryl</Name>
         *         <HighValueCustomer>True</HighValueCustomer>
         *         <Orders>
         *           <Order>
         *             <ProductDescription>Unicycle</ProductDescription>
         *             <Quantity>3</Quantity>
         *             <OrderDate>9/5/2001</OrderDate>
         *           </Order>
         *           <Order>
         *             <ProductDescription>Tricycle</ProductDescription>
         *             <Quantity>3</Quantity>
         *             <OrderDate>8/6/2000</OrderDate>
         *           </Order>
         *         </Orders>
         *       </Customer>
         *     </Customers>
         *   </DocumentGenerationData>
         * </GenerateDocuments>
         */

        /*
         * Return message:
         * 
         * Message label: DocumentGenerationComplete
         * 
         * <DocumentsGenerated>
         *   <Document Name="48FCE74D-E9E1-49DB-ABCD-629196B6AEE5.docx" />
         * </DocumentsGenerated>
         */
        private static void MessageLoop()
        {
            System.Messaging.Message generateDocumentsMessage;
            string generateDocumentsMessageText;

            while (true)
            {
                generateDocumentsMessage = GenerateDocumentsQueue.Receive();
                generateDocumentsMessage.Formatter = new XmlMessageFormatter(new String[] { "System.String,mscorlib" });
                generateDocumentsMessageText = generateDocumentsMessage.Body.ToString();

                if (generateDocumentsMessage.Label == "GenerateDocuments")
                {
                    var generateDocumentsArguments = XElement.Parse(generateDocumentsMessageText);

                    // If you want to validate the XML in the message, this is the place to do it.

                    var returnQueueName = (string)generateDocumentsArguments
                        .Element("ReturnQueue")
                        .Attribute("Name");
                    var templateDocumentName = (string)generateDocumentsArguments
                        .Element("TemplateDocument")
                        .Attribute("Name");
                    var generatedDocumentsDirectory = (string)generateDocumentsArguments
                        .Element("GeneratedDocumentsDirectory")
                        .Attribute("Name");
                    List<string> generatedDocumentsList = DocumentGenerator
                        .GenerateDocuments(
                            new FileInfo(templateDocumentName), 
                            generateDocumentsArguments.Element("DocumentGenerationData").Elements().First(), 
                            new DirectoryInfo(generatedDocumentsDirectory));
                    foreach (var doc in generatedDocumentsList)
                        Console.WriteLine("Daemon generated {0}", doc);

                    MessageQueue returnQueue = new MessageQueue(returnQueueName, false, false, QueueAccessMode.Send);
                    Message docsCompleteMessage = new Message();
                    var cmsg = new XElement("DocumentsGenerated",
                        generatedDocumentsList.Select(dn => new XElement("Document",
                            new XAttribute("Name", dn))));
                    docsCompleteMessage.Body = cmsg.ToString(SaveOptions.DisableFormatting);
                    docsCompleteMessage.Label = "DocumentGenerationComplete";
                    returnQueue.Send(docsCompleteMessage);

                    continue;
                }

                throw new DocumentGenerationDaemonException("Internal error: Should never reach here");
            }
        }

        private static void Initialize()
        {
            if (MessageQueue.Exists(GenerateDocumentsQueueLocalName))
                MessageQueue.Delete(GenerateDocumentsQueueLocalName);
            GenerateDocumentsQueue = MessageQueue.Create(GenerateDocumentsQueueLocalName, false);
        }
    }

    class DocumentGenerator
    {
        private static XNamespace w = "http://schemas.openxmlformats.org/wordprocessingml/2006/main";
        private static string SelectDocuments { get; set; }

        private static XElement GetContentControlByTag(XContainer ancestor, string tag)
        {
            return ancestor.Descendants(w + "sdt")
                .Where(e => e.Elements(w + "sdtPr")
                    .Elements(w + "alias")
                    .Attributes(w + "val")
                    .FirstOrDefault().Value == tag)
                .FirstOrDefault();
        }

        private static string GetContentControlContents(XElement contentControl)
        {
            return contentControl.Element(w + "sdtContent")
                .Descendants(w + "t")
                .Select(t => (string)t)
                .StringConcatenate();
        }

        private static object Transform(XNode node, XElement document)
        {
            XElement element = node as XElement;
            if (element != null)
            {
                if (element.Name == w + "sdt")
                {
                    string tag = element.Elements(w + "sdtPr")
                        .Elements(w + "alias")
                        .Attributes(w + "val")
                        .FirstOrDefault()
                        .Value;
                    if (tag == "Config") return null;
                    if (tag == "SelectValue")
                    {
                        XElement run = element.Element(w + "sdtContent").Element(w + "r");
                        string valueSelector = GetContentControlContents(element);
                        string newValue = document.XPathSelectElement(valueSelector).Value;
                        return (new XElement(w + "r",
                            run.Elements().Where(e => e.Name != w + "t"),
                            new XElement(w + "t", newValue)));
                    }
                    if (tag == "Repeat")
                    {
                        XElement contentContentControl = GetContentControlByTag(element, "Content");
                        XElement selectRepeatingDataContentControl =
                            GetContentControlByTag(element, "SelectRepeatingData");
                        string selector = GetContentControlContents(selectRepeatingDataContentControl);
                        var repeatingData = document.XPathSelectElements(selector);
                        var newContent = repeatingData.Select(d =>
                        {
                            var content = contentContentControl.Element(w + "sdtContent")
                                .Elements().Select(e => Transform(e, d)).ToList();
                            return content;
                        })
                            .ToList();
                        return newContent;
                    }
                    if (tag == "Table")
                    {
                        XElement content = GetContentControlByTag(element, "Content");
                        XElement selectRowsContentControl =
                            GetContentControlByTag(element, "SelectRows");
                        string selector = GetContentControlContents(selectRowsContentControl);
                        var tableData = document.XPathSelectElements(selector);
                        XElement table = content.Descendants(w + "tbl").FirstOrDefault();
                        XElement protoRow = table.Elements(w + "tr").Skip(1).FirstOrDefault();
                        XElement newTable = new XElement(w + "tbl",
                            table.Elements().Where(e => e.Name != w + "tr"),
                            table.Elements(w + "tr").FirstOrDefault(),
                            tableData.Select(d =>
                                new XElement(w + "tr",
                                    protoRow.Elements().Where(r => r.Name != w + "tc"),
                                    protoRow.Elements(w + "tc")
                                        .Select(tc =>
                                        {
                                            XElement paragraph = tc.Elements(w + "p")
                                                .FirstOrDefault();
                                            XElement run = paragraph.Elements(w + "r")
                                                .FirstOrDefault();
                                            string cellSelector = paragraph.Value;
                                            string cellData =
                                                d.XPathSelectElement(cellSelector).Value;
                                            XElement newCell = new XElement(w + "tc",
                                                tc.Elements().Where(z => z.Name != w + "p"),
                                                new XElement(w + "p",
                                                    paragraph.Elements().Where(z1 => z1.Name != w + "r"),
                                                    new XElement(w + "r",
                                                        run.Elements().Where(z2 => z2.Name != w + "t"),
                                                        new XElement(w + "t", cellData))));
                                            return newCell;
                                        }))));
                        return newTable;
                    }
                    if (tag == "Conditional")
                    {
                        XElement selectTestValueContentControl =
                            GetContentControlByTag(element, "SelectTestValue");
                        string selectTestValue =
                            GetContentControlContents(selectTestValueContentControl);
                        string match = GetContentControlContents(
                            GetContentControlByTag(element, "Match"));
                        XElement contentContentControl =
                            GetContentControlByTag(element, "Content");
                        string value =
                            document.XPathSelectElement(selectTestValue).Value;
                        if (value == match)
                        {
                            var content = contentContentControl.Element(w + "sdtContent")
                                .Elements().Select(e => Transform(e, document));
                            return content;
                        }
                        else
                            return null;
                    }
                }
                return new XElement(element.Name,
                    element.Attributes(),
                    element.Nodes().Select(n => Transform(n, document)));
            }
            return node;
        }

        private static void GenerateDocument(FileInfo templateDocument, XElement documentGenerationInfo, string generatedDocumentFileName)
        {
            byte[] byteArray = File.ReadAllBytes(templateDocument.FullName);
            using (MemoryStream mem = new MemoryStream())
            {
                mem.Write(byteArray, 0, (int)byteArray.Length);
                using (WordprocessingDocument wordDoc =
                    WordprocessingDocument.Open(mem, true))
                {
                    XDocument xDoc = wordDoc.MainDocumentPart.GetXDocument();
                    XElement newRootElement = (XElement)Transform(
                        wordDoc.MainDocumentPart.GetXDocument().Root, documentGenerationInfo);
                    xDoc.Elements().First().ReplaceWith(newRootElement);
                    wordDoc.MainDocumentPart.PutXDocument();
                }
                using (FileStream fileStream = new FileStream(generatedDocumentFileName,
                    System.IO.FileMode.CreateNew))
                    mem.WriteTo(fileStream);
            }
        }

        public static List<string> GenerateDocuments(FileInfo templateDocumentFileInfo, XElement documentGenerationData, DirectoryInfo generatedDocumentsDirectory)
        {
            List<string> generatedDocuments = new List<string>();
            using (WordprocessingDocument templateDocument = WordprocessingDocument.Open(templateDocumentFileInfo.FullName, false))
            {
                XDocument mainDocumentPart = templateDocument.MainDocumentPart.GetXDocument();
                XElement configContentControl = GetContentControlByTag(mainDocumentPart, "Config");
                XElement config = XElement.Parse(configContentControl.Element(w + "sdtContent")
                    .Descendants(w + "t")
                    .Select(t => (string)t)
                    .StringConcatenate());
                SelectDocuments = config.Element("SelectDocuments").Value;
                var selectedDocuments = documentGenerationData.XPathSelectElements(SelectDocuments);
                int i = 0;
                foreach (var documentInfo in selectedDocuments)
                {
                    string generatedDocumentFileName = generatedDocumentsDirectory + Guid.NewGuid().ToString() + ".docx";
                    generatedDocuments.Add(generatedDocumentFileName);
                    GenerateDocument(templateDocumentFileInfo, documentInfo, generatedDocumentFileName);
                    if (++i % 100 == 0) Console.WriteLine(i);
                }
            }
            return generatedDocuments;
        }
    }

    public static class LocalExtensions
    {
        public static string StringConcatenate(this IEnumerable<string> source)
        {
            StringBuilder sb = new StringBuilder();
            foreach (string item in source) sb.Append(item);
            return sb.ToString();
        }

        public static XDocument GetXDocument(this OpenXmlPart part)
        {
            XDocument xdoc = part.Annotation<XDocument>();
            if (xdoc != null) return xdoc;
            using (StreamReader sr = new StreamReader(part.GetStream()))
            using (XmlReader xr = XmlReader.Create(sr))
                xdoc = XDocument.Load(xr);
            part.AddAnnotation(xdoc);
            return xdoc;
        }

        public static void PutXDocument(this OpenXmlPart part)
        {
            XDocument xdoc = part.GetXDocument();
            if (xdoc != null)
                using (XmlWriter xw = XmlWriter.Create(part.GetStream
                  (FileMode.Create, FileAccess.Write)))
                    xdoc.Save(xw);
        }
    }

    public class DocumentGenerationDaemonException : Exception
    {
        public DocumentGenerationDaemonException(string message) : base(message) { }
    }
}
